#include "mod_dav_dpm.h"
#include <dpm_api.h>
#include <openssl/evp.h>
#include <openssl/sha.h>
#include <serrno.h>
#include <apr_file_io.h>

/* Not nice to have globals, but it is internal to this module... */
static char file_password[256] = {'\0'};

/**
 * Loads the password from a file
 * @param p    The pool
 * @param path The file where the password is
 * @return NULL on success
 */
static dav_error *load_file_password(apr_pool_t *p, const char *path)
{
  apr_file_t *file;
  apr_size_t nbytes = sizeof(file_password);

  if (apr_file_open(&file, path, APR_READ, APR_OS_DEFAULT, p) != APR_SUCCESS) {
    return dav_dpm_new_error(p, errno, HTTP_INTERNAL_SERVER_ERROR,
                             "Could not load the password from the file");
  }

  apr_file_read(file, file_password, &nbytes);
  apr_file_close(file);

  file_password[nbytes] = '\0';

  if (nbytes == 0) {
    return dav_dpm_new_error(p, EINVAL, 0, "Empty passwords forbidden!!");
  }

  return NULL;
}

/* Generation: DPM token implementation */
dav_error *dpm_generate_token(apr_pool_t *p, char *token, size_t st,
                              const char *sfn, const char *path,
                              const char *r_token, const char *extra)
{
  strncpy(token, r_token, st);
  return NULL;
}

/* Validation: DPM token implementation */
dav_error *dpm_validate_token(apr_pool_t *p, const char *token,
                              const char *sfn, const char *path,
                              const char *extra)
{
  int                       nreplies;
  struct dpm_getfilestatus *statuses;
  struct physical_loc       location;

  /* Get information about the token */
  if (dpm_getstatus_getreq((char*)token, 1, (char**)&sfn, &nreplies, &statuses) < 0 || nreplies < 1) {
    return dav_dpm_new_error(p, serrno, 0, "Could not validate the token");
  }
 
  /* Check that everything fits */
  dav_dpm_split_turl(p, statuses[0].turl, &location);
  if (strcmp(path, location.path) != 0) {
    dpm_free_gfilest(nreplies, statuses);
    return dav_dpm_new_error(p, EACCES, HTTP_FORBIDDEN, "The token does not validate!");
  }

  dpm_free_gfilest(nreplies, statuses);

  return NULL;
}

/* Generation: Pass phrase implementation */
dav_error *pass_generate_token(apr_pool_t *p, char *token, size_t st,
                               const char *sfn, const char *path,
                               const char *r_token, const char *extra)
{
  unsigned char  append_buffer[1024], cipher_buffer[1024];
  unsigned char  sha[SHA_DIGEST_LENGTH];
  EVP_CIPHER_CTX cipher_ctx;
  int            outl, lastl, n, i;

  n = snprintf((char*)append_buffer, sizeof(append_buffer), "%s:%s", sfn, path);

  EVP_EncryptInit(&cipher_ctx, EVP_aes_128_ecb(), (unsigned const char*)extra, NULL);
  EVP_EncryptUpdate(&cipher_ctx, cipher_buffer, &outl, append_buffer, n);
  EVP_EncryptFinal(&cipher_ctx, cipher_buffer + outl, &lastl);
  outl += lastl;

  SHA1(cipher_buffer, outl, sha);

  for (i = 0; i < SHA_DIGEST_LENGTH; ++i) {
    sprintf(token + i * 2, "%02x", sha[i]);
  }

  return NULL;
}

/* Validation: Pass phrase implementation */
dav_error *pass_validate_token(apr_pool_t *p, const char *token,
                               const char *sfn, const char *path,
                               const char *extra)
{
  char           buffer[1024];
  unsigned char  cipher_buffer[1024];
  unsigned char  sha[SHA_DIGEST_LENGTH];
  EVP_CIPHER_CTX cipher_ctx;
  int            outl, lastl, n, i;

  n = snprintf((char*)buffer, sizeof(buffer), "%s:%s", sfn, path);

  EVP_EncryptInit(&cipher_ctx, EVP_aes_128_ecb(), (unsigned const char*)extra, NULL);
  EVP_EncryptUpdate(&cipher_ctx, cipher_buffer, &outl, (unsigned char*)buffer, n);
  EVP_EncryptFinal(&cipher_ctx, cipher_buffer + outl, &lastl);
  outl += lastl;

  SHA1(cipher_buffer, outl, sha);

  for (i = 0; i < SHA_DIGEST_LENGTH; ++i) {
    sprintf(buffer + i * 2, "%02x", sha[i]);
  }

  if (strcmp(buffer, token) != 0) {
    return dav_dpm_new_error(p, EACCES, HTTP_FORBIDDEN, "The token does not validate!");
  }

  return NULL;
}

/* Generation: Pass file implementation */
dav_error *file_generate_token(apr_pool_t *p, char *token, size_t st,
                               const char *sfn, const char *path,
                               const char *r_token, const char *extra)
{
  dav_error *err;
  /* Maybe need to load the password! */
  if (file_password[0] == '\0') {
    err = load_file_password(p, extra);
    if (err)
      return err;
  }
  /* Delegate on pass_generate_token */
  return pass_generate_token(p, token, st, sfn, path, r_token, file_password);
}

/* Validation: Pass file implementation */
dav_error *file_validate_token(apr_pool_t *p, const char *token,
                               const char *sfn, const char *path,
                               const char *extra)
{
  dav_error *err = NULL;
  
  /* Maybe need to load the password! */
  if (file_password[0] == '\0') {
    err = load_file_password(p, extra);
    if (err)
      return err;
  }

  /* Delegate: First attemp */
  err = pass_validate_token(p, token, sfn, path, file_password);

  if (err) {
    /* Reload password from file and second attemp */
    err = load_file_password(p, extra);
    if (!err)
      err = pass_validate_token(p, token, sfn, path, file_password);
  }

  return err;
}
